==> csircd-1.13.tar.gz: csircd-1.13/source/ircdcore.c <==
/*
**   IRC - Internet Relay Chat, source/ircdcore.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "setup.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <signal.h>
#include <syslog.h>
#include <time.h>
#include <sys/time.h> 
#include <sys/resource.h>
#ifdef HAVE_DLOPEN
#include <dlfcn.h>
#endif
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include "config.h"

#ifdef HAVE_DLOPEN

typedef struct _ircd_symbol
{
	char	*name;
#define IRCD_SYMBOL_FLAGS_NOTREQUIRED	0x001
	int		flags;
	void	*saddr;
} IRCD_SYMBOL;

static IRCD_SYMBOL Symbols[] =
{
#define IRCD_SYMBOL_MAIN_LOOP				0
	{	"_ircd_main_loop",			0,		NULL },
#define IRCD_SYMBOL_IRCDSO_VERSIONSTRING	1
	{	"_ircdso_versionstring",	0,		NULL },
#define IRCD_SYMBOL_IRCDSO_VERSION			2
	{	"_ircdso_version",			0,		NULL },
	{	NULL,						0,		NULL }
};

#else

extern int _ircd_main_loop();

#endif

static void	*Data = NULL;
static int	Data_size = 0;
static int	Data_format_version = 0;
static char	*Ircdcore_versionstring = "1.13-REL";
#ifdef HAVE_DLOPEN
static int	Ircdcore_version = 2;
#endif

typedef struct _data
{
	char	*Conffile;
} IRCD_DATA;


static void usage(void);
static int is_numeric_str(char *str);

static void usage(void)
{
	fprintf(stderr, "usage: ircd [-l] [-i] [-n] [-u user/uid] [-g group/gid] [-t chrootdir] [-p path/to/csircd.so] <config file>\n");
	exit(1);
}

static int is_numeric_str(char *str)
{
	char	*ptr;

	if (!str)
		return 0;
	for(ptr=str;*ptr;ptr++)
	{
		if ((*ptr < '0') || (*ptr > '9'))
			return 0;
	}
	return 1;
}

int main(int argc, char *argv[])
{
	int				firsttime = 1;
#ifdef HAVE_DLOPEN
	void			*vptr;
	IRCD_SYMBOL		*sptr;
#endif
	int				ret;
	int				ch;
	uid_t			uid = 0;
	gid_t			gid = 0;
	char			*chrootdir = NULL;
	int				doversion = 0;
	int				nofork = 0;
	int				(*mainloop)();
	int				nounlimit = 0;
	char			*pathtoircdso = "./csircd.so";

	assert(chrootdir == NULL); /* hack to get gcc to link in libgcc.a which is needed
					for assert()s in csircd.so */

	while ((ch = getopt(argc, argv, "u:g:t:linvh")) != EOF)
	{
		switch(ch)
		{
			case 'u':
			{
				struct passwd *pw;

				if (is_numeric_str(optarg))
					pw = getpwuid(atoi(optarg));
				else
					pw = getpwnam(optarg);

				if (!pw)
				{
					fprintf(stderr, "uid \"%s\" doesn't exist\n", optarg);
					exit(-1);
				}

				uid = pw->pw_uid;

				if (!gid)
				{
					struct group *gr;

					gr = getgrgid(pw->pw_gid);
					if (gr)
						gid = gr->gr_gid;
				}
			}
				break;
			case 'g':
			{
				struct group *gr;

				if (is_numeric_str(optarg))
					gr = getgrgid(atoi(optarg));
				else
					gr = getgrnam(optarg);

				if (!gr)
				{
					fprintf(stderr, "gid \"%s\" doesn't exist\n", optarg);
					exit(-1);
				}

				gid = gr->gr_gid;

			}
				break;
			case 't':
				if (geteuid())
				{
					fprintf(stderr, "must be root to chroot()\n");
					exit(-1);
				}
				else
					chrootdir = optarg;
				break;
			case 'l':
				nounlimit = 0;
				break;
			case 'i':
				nounlimit = 1;
				break;
			case 'n':
				nofork = 1;
				break;
			case 'v':
				doversion = 1;
				break;
			case 'p':
				if (!optarg || !*optarg)
				{
					fprintf(stderr, "Invalid library name\n");
					exit(-1);
				}
				pathtoircdso = optarg;
				break;
			case 'h':
			default:
				usage();
		}
	}

	argc -= optind;
	argv += optind;

	if (argc < 1)
		usage();

	if (!nounlimit)
	{
		struct rlimit limits[8];
		int i;

		if (!geteuid())
		{
			for(i=0;i<8;++i)
			{
				limits[i].rlim_cur = RLIM_INFINITY;
				limits[i].rlim_max = RLIM_INFINITY;
			}
		}
		else
		{
			getrlimit(RLIMIT_CORE, limits+0);
			getrlimit(RLIMIT_CPU, limits+1);
			getrlimit(RLIMIT_DATA, limits+2);
			getrlimit(RLIMIT_FSIZE, limits+3);
#ifdef RLIMIT_MEMLOCK
			getrlimit(RLIMIT_MEMLOCK, limits+4);
#endif
			getrlimit(RLIMIT_NOFILE, limits+5);
#ifdef RLIMIT_RSS
			getrlimit(RLIMIT_RSS, limits+6);
#endif
			getrlimit(RLIMIT_STACK, limits+7);

			for(i=0;i<8;++i)
				limits[i].rlim_cur = limits[i].rlim_max;
		}

		setrlimit(RLIMIT_CORE, limits+0);
		setrlimit(RLIMIT_CPU, limits+1);
		setrlimit(RLIMIT_DATA, limits+2);
		setrlimit(RLIMIT_FSIZE, limits+3);
#ifdef RLIMIT_MEMLOCK
		setrlimit(RLIMIT_MEMLOCK, limits+4);
#endif
		setrlimit(RLIMIT_NOFILE, limits+5);
#ifdef RLIMIT_RSS
		setrlimit(RLIMIT_RSS, limits+6);
#endif
		setrlimit(RLIMIT_STACK, limits+7);
	}

	if (chrootdir)
	{
		if (chdir(chrootdir) < 0)
		{
			fprintf(stderr, "chdir to \"%s\" for chroot failed: %i\n",
					chrootdir, -errno);
			exit(-1);
		}
		if (chroot(".") < 0)
		{
			fprintf(stderr, "chroot to \"%s\" failed: %i\n",
					chrootdir, -errno);
			exit(-1);
		}
	}

	if (gid)
	{
		if (setgid(gid) < 0)
		{
			fprintf(stderr, "setgid(%u) failed: %i\n", (u_int)gid, -errno);
			exit(-1);
		}
		syslog(LOG_NOTICE, "gid changed to %u", gid);
	}

	if (uid)
	{
		if (setuid(uid) < 0)
		{
			fprintf(stderr, "setuid(%u) failed: %i\n", (u_int)uid, -errno);
			exit(-1);
		}
		syslog(LOG_NOTICE, "uid changed to %u", gid);
	}

	openlog("csircd", LOG_PID|LOG_NDELAY, LOG_DAEMON);

	signal(SIGPIPE, SIG_IGN);
	signal(SIGTRAP, SIG_IGN);
	signal(SIGXCPU, SIG_IGN);
	signal(SIGXFSZ, SIG_IGN);
	signal(SIGHUP, SIG_IGN);

	Data = (char *)malloc(sizeof(IRCD_DATA));
	Data_size = sizeof(IRCD_DATA);
	((IRCD_DATA *)Data)->Conffile = argv[0];

	for(;;)
	{
#ifdef HAVE_DLOPEN
		vptr = dlopen(pathtoircdso, RTLD_NOW);
		if (!vptr)
		{
			const char *err = dlerror();

			syslog(LOG_ERR, "can't dlopen %s: %s",
					pathtoircdso, err);
			if (firsttime)
			{
				fprintf(stderr, "can't dlopen %s: %s\n",
					pathtoircdso, err);
				exit(-1);
			}
		}

		for(sptr=Symbols;sptr->name;sptr++)
		{
			sptr->saddr = dlsym(vptr, sptr->name);
			if (!sptr->saddr && !(sptr->flags & IRCD_SYMBOL_FLAGS_NOTREQUIRED))
			{
				fprintf(stderr,
					"required symbol \"%s\" is missing from %s...\n",
					sptr->name, pathtoircdso);
				syslog(LOG_ERR,
					"required symbol \"%s\" is missing from %s...",
					sptr->name, pathtoircdso);
				break;
			}
		}

		if (sptr->name)
		{
			dlclose(vptr);
			if (firsttime)
				exit(-1);
			sleep(1);
			continue;
		}

		if ((ret=*((int *)Symbols[IRCD_SYMBOL_IRCDSO_VERSION].saddr)) > Ircdcore_version)
		{
			syslog(LOG_ERR, "csircdso arg version %d > ircdcore arg version %d",
					ret, Ircdcore_version);
			fprintf(stderr, "csircdso arg version %d > ircdcore arg version %d\n",
					ret, Ircdcore_version);
			if (firsttime)
				exit(-1);
		}

		syslog(LOG_NOTICE, "version %s of %s has been loaded",
				(char *)Symbols[IRCD_SYMBOL_IRCDSO_VERSIONSTRING].saddr,
				pathtoircdso);
		if (doversion)
		{
			fprintf(stderr, "%s version %s\n",
				pathtoircdso,
				(char *)Symbols[IRCD_SYMBOL_IRCDSO_VERSIONSTRING].saddr);
			dlclose(vptr);
			exit(0);
		}

		mainloop = Symbols[IRCD_SYMBOL_MAIN_LOOP].saddr;

#else

		mainloop = _ircd_main_loop;

#endif

		if (!nofork)
		{
			ret = fork();
			if (ret < 0)
				fprintf(stderr, "Couldn't fork into the background, running in foreground.\n");
			else if (ret)
				exit(0);
			nofork = 1;
		}

		ret = mainloop(
					firsttime,
					Ircdcore_versionstring,
					&Data_format_version,
					&Data,
					&Data_size);

#ifdef HAVE_DLOPEN
		dlclose(vptr);
#endif
		if (ret < 0)
		{
			sleep(1);
			continue;
		}
		firsttime = 0;
	}

	/* NOT REACHED */
	exit(0);
}

==> csircd-1.14.tar.gz: csircd-1.14/source/ircdcore.c <==
/*
**   IRC - Internet Relay Chat, source/ircdcore.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "setup.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <signal.h>
#include <syslog.h>
#include <time.h>
#include <sys/time.h> 
#include <sys/resource.h>
#ifdef HAVE_DLOPEN
#include <dlfcn.h>
#endif
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include "config.h"
#include "version.h"

#ifdef HAVE_DLOPEN

typedef struct _ircd_symbol
{
	char	*name;
#define IRCD_SYMBOL_FLAGS_NOTREQUIRED	0x001
	int		flags;
	void	*saddr;
} IRCD_SYMBOL;

static IRCD_SYMBOL Symbols[] =
{
#define IRCD_SYMBOL_MAIN_LOOP				0
	{	"_ircd_main_loop",			0,		NULL },
#define IRCD_SYMBOL_IRCDSO_VERSIONSTRING	1
	{	"_ircdso_versionstring",	0,		NULL },
#define IRCD_SYMBOL_IRCDSO_VERSION			2
	{	"_ircdso_version",			0,		NULL },
	{	NULL,						0,		NULL }
};

#else

extern int _ircd_main_loop();

#endif

static void	*Data = NULL;
static int	Data_size = 0;
static int	Data_format_version = CSIRCD_DATA_FORMAT;
static char	*Ircdcore_versionstring = VERSION;

typedef struct _data
{
	char	**argv;
	char	*Conffile;
} IRCD_DATA;


static void usage(void);
static int is_numeric_str(char *str);

static void usage(void)
{
	fprintf(stderr, "Usage: csircd [-l] [-i] [-n] [-u user/uid] [-g group/gid] [-t chrootdir] [-p path/to/csircd.so] <config file>\n");
	exit(1);
}

static int is_numeric_str(char *str)
{
	char	*ptr;

	if (!str)
		return 0;
	for(ptr=str;*ptr;ptr++)
	{
		if ((*ptr < '0') || (*ptr > '9'))
			return 0;
	}
	return 1;
}

int main(int argc, char *argv[])
{
	int				firsttime = 1;
#ifdef HAVE_DLOPEN
	void			*vptr;
	IRCD_SYMBOL		*sptr;
#endif
	int				ret;
	int				ch;
	uid_t			uid = 0;
	gid_t			gid = 0;
	char			*chrootdir = NULL;
	char			*ptr;
	int				nofork = 0;
	int				(*mainloop)();
	int				nounlimit = 0;
	char			*pathtoircdso = CSIRCDSO_DIR "/csircd.so";
	char			**argv_orig = argv;

	while ((ch = getopt(argc, argv, "p:u:g:T:t:linvh")) != EOF)
	{
		switch(ch)
		{
			case 'u':
			{
				struct passwd *pw;

				if (is_numeric_str(optarg))
					pw = getpwuid(atoi(optarg));
				else
					pw = getpwnam(optarg);

				if (!pw)
				{
					fprintf(stderr, "uid \"%s\" doesn't exist\n", optarg);
					exit(-1);
				}

				uid = pw->pw_uid;

				if (!gid)
				{
					struct group *gr;

					gr = getgrgid(pw->pw_gid);
					if (gr)
						gid = gr->gr_gid;
				}
			}
				break;
			case 'g':
			{
				struct group *gr;

				if (is_numeric_str(optarg))
					gr = getgrgid(atoi(optarg));
				else
					gr = getgrnam(optarg);

				if (!gr)
				{
					fprintf(stderr, "gid \"%s\" doesn't exist\n", optarg);
					exit(-1);
				}

				gid = gr->gr_gid;

			}
				break;
			case 'T':
				/* 't's are converted to Ts and ignored on /restart
				**  since we've already chrooted
				*/
				break;
			case 't':
				if (geteuid())
				{
					fprintf(stderr, "must be root to chroot()\n");
					exit(-1);
				}
				else
					chrootdir = optarg;

				ptr = argv[optind-2];
				while(*ptr)
				{
					if (*ptr == 't')
					{
						*ptr = 'T';
						break;
					}
					ptr++;
				}
				break;
			case 'l':
				nounlimit = 0;
				break;
			case 'i':
				nounlimit = 1;
				break;
			case 'n':
				nofork = 1;
				break;
			case 'v':
				fprintf(stderr, "%s version %s (data format version %d)\n",
					argv[0],
					Ircdcore_versionstring,
					Data_format_version);
#ifdef HAVE_DLOPEN
				vptr = dlopen(pathtoircdso, RTLD_NOW);
				if (!vptr)
				{
					const char *err = dlerror();

					fprintf(stderr, "can't dlopen %s: %s\n",
							pathtoircdso, err);
					exit(-1);
				}

				sptr = Symbols + IRCD_SYMBOL_IRCDSO_VERSIONSTRING;
				sptr->saddr = dlsym(vptr, sptr->name);
				if (!sptr->saddr)
				{
					fprintf(stderr,
						"symbol \"%s\" is missing from %s...\n",
						sptr->name, pathtoircdso);
				}
				else
				{
					fprintf(stderr, "%s version %s\n",
						pathtoircdso,
						(char *)sptr->saddr);
				}

				sptr = Symbols + IRCD_SYMBOL_IRCDSO_VERSION;
				sptr->saddr = dlsym(vptr, sptr->name);
				if (!sptr->saddr)
				{
					fprintf(stderr,
						"symbol \"%s\" is missing from %s...\n",
						sptr->name, pathtoircdso);
				}
				else
				{
					fprintf(stderr, "%s data format version %d\n",
						pathtoircdso,
						*((int *)sptr->saddr));
				}
#endif
				exit(0);
				break;
			case 'p':
				if (!optarg || !*optarg)
				{
					fprintf(stderr, "Invalid library name\n");
					exit(-1);
				}
				pathtoircdso = optarg;
				break;
			case 'h':
			default:
				usage();
		}
	}

	argc -= optind;
	argv += optind;

	if (argc < 1)
	{
		usage();
	}

	if (!nounlimit)
	{
		struct rlimit limits[8];
		int i;

		if (!geteuid())
		{
			for(i=0;i<8;++i)
			{
				limits[i].rlim_cur = RLIM_INFINITY;
				limits[i].rlim_max = RLIM_INFINITY;
			}
		}
		else
		{
			getrlimit(RLIMIT_CORE, limits+0);
			getrlimit(RLIMIT_CPU, limits+1);
			getrlimit(RLIMIT_DATA, limits+2);
			getrlimit(RLIMIT_FSIZE, limits+3);
#ifdef RLIMIT_MEMLOCK
			getrlimit(RLIMIT_MEMLOCK, limits+4);
#endif
			getrlimit(RLIMIT_NOFILE, limits+5);
#ifdef RLIMIT_RSS
			getrlimit(RLIMIT_RSS, limits+6);
#endif
			getrlimit(RLIMIT_STACK, limits+7);

			for(i=0;i<8;++i)
				limits[i].rlim_cur = limits[i].rlim_max;
		}

		setrlimit(RLIMIT_CORE, limits+0);
		setrlimit(RLIMIT_CPU, limits+1);
		setrlimit(RLIMIT_DATA, limits+2);
		setrlimit(RLIMIT_FSIZE, limits+3);
#ifdef RLIMIT_MEMLOCK
		setrlimit(RLIMIT_MEMLOCK, limits+4);
#endif
		setrlimit(RLIMIT_NOFILE, limits+5);
#ifdef RLIMIT_RSS
		setrlimit(RLIMIT_RSS, limits+6);
#endif
		setrlimit(RLIMIT_STACK, limits+7);
	}

	if (chrootdir)
	{
		if (chdir(chrootdir) < 0)
		{
			fprintf(stderr, "chdir to \"%s\" for chroot failed: %i\n",
					chrootdir, -errno);
			exit(-1);
		}
		if (chroot(".") < 0)
		{
			fprintf(stderr, "chroot to \"%s\" failed: %i\n",
					chrootdir, -errno);
			exit(-1);
		}
	}

	if (gid)
	{
		if (setgid(gid) < 0)
		{
			fprintf(stderr, "setgid(%u) failed: %i\n", (u_int)gid, -errno);
			exit(-1);
		}
		syslog(LOG_NOTICE, "gid changed to %u", gid);
	}

	if (uid)
	{
		if (setuid(uid) < 0)
		{
			fprintf(stderr, "setuid(%u) failed: %i\n", (u_int)uid, -errno);
			exit(-1);
		}
		syslog(LOG_NOTICE, "uid changed to %u", gid);
	}

	openlog("csircd", LOG_PID|LOG_NDELAY, LOG_DAEMON);

	signal(SIGPIPE, SIG_IGN);
	signal(SIGTRAP, SIG_IGN);
	signal(SIGXCPU, SIG_IGN);
	signal(SIGXFSZ, SIG_IGN);
	signal(SIGHUP, SIG_IGN);

	argv_orig[0] = CSIRCD_DIR "/csircd";

	Data = (char *)malloc(sizeof(IRCD_DATA));
	Data_size = sizeof(IRCD_DATA);
	((IRCD_DATA *)Data)->Conffile = argv[0];
	((IRCD_DATA *)Data)->argv = argv_orig;

	for(;;)
	{
#ifdef HAVE_DLOPEN
		vptr = dlopen(pathtoircdso, RTLD_NOW);
		if (!vptr)
		{
			const char *err = dlerror();

			syslog(LOG_ERR, "can't dlopen %s: %s",
					pathtoircdso, err);
			if (firsttime)
			{
				fprintf(stderr, "can't dlopen %s: %s\n",
					pathtoircdso, err);
				exit(-1);
			}
		}

		for(sptr=Symbols;sptr->name;sptr++)
		{
			sptr->saddr = dlsym(vptr, sptr->name);
			if (!sptr->saddr && !(sptr->flags & IRCD_SYMBOL_FLAGS_NOTREQUIRED))
			{
				fprintf(stderr,
					"required symbol \"%s\" is missing from %s...\n",
					sptr->name, pathtoircdso);
				syslog(LOG_ERR,
					"required symbol \"%s\" is missing from %s...",
					sptr->name, pathtoircdso);
				break;
			}
		}

		if (sptr->name)
		{
			dlclose(vptr);
			if (firsttime)
				exit(-1);
			sleep(1);
			continue;
		}

		assert(*((int *)Symbols[IRCD_SYMBOL_IRCDSO_VERSION].saddr) > 0);

#if 0
csircd.so checks this in main_loop...

		if ((ret=*((int *)Symbols[IRCD_SYMBOL_IRCDSO_VERSION].saddr)) > Ircdcore_version)
		{
			syslog(LOG_ERR, "csircdso arg version %d > ircdcore arg version %d",
					ret, Ircdcore_version);
			fprintf(stderr, "csircdso arg version %d > ircdcore arg version %d\n",
					ret, Ircdcore_version);
			dlclose(vptr);
			if (firsttime)
				exit(-1);
			sleep(1);
			continue;
		}


#endif

		syslog(LOG_NOTICE, "version %s of %s has been loaded (format version %d)",
				(char *)Symbols[IRCD_SYMBOL_IRCDSO_VERSIONSTRING].saddr,
				pathtoircdso, *((int *)Symbols[IRCD_SYMBOL_IRCDSO_VERSION].saddr));

		mainloop = Symbols[IRCD_SYMBOL_MAIN_LOOP].saddr;

#else

		mainloop = _ircd_main_loop;

#endif

		if (!nofork)
		{
			ret = fork();
			if (ret < 0)
				fprintf(stderr, "Couldn't fork into the background, running in foreground.\n");
			else if (ret)
				exit(0);
			nofork = 1;
		}

		ret = mainloop(
					firsttime,
					Ircdcore_versionstring,
					&Data_format_version,
					&Data,
					&Data_size);

#ifdef HAVE_DLOPEN
		dlclose(vptr);
#endif
		if (ret < 0)
		{
			sleep(1);
			continue;
		}
		firsttime = 0;
	}

	/* NOT REACHED */
	exit(0);
}
